#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import http.client
import uuid

from keystone.common import provider_api
import keystone.conf
from keystone.tests.common import auth as common_auth
from keystone.tests import unit
from keystone.tests.unit import base_classes
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import temporaryfile

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class _UserAccessRuleTests:
    """Test cases for anyone that has a valid user token."""

    def test_user_can_get_their_access_rules(self):
        access_rule_id = uuid.uuid4().hex
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': self.user_id,
            'project_id': self.project_id,
            'secret': uuid.uuid4().hex,
            'access_rules': [
                {
                    'id': access_rule_id,
                    'service': uuid.uuid4().hex,
                    'path': uuid.uuid4().hex,
                    'method': uuid.uuid4().hex[16:],
                }
            ],
        }
        PROVIDERS.application_credential_api.create_application_credential(
            app_cred
        )
        with self.test_client() as c:
            path = '/v3/users/{}/access_rules/{}'.format(
                self.user_id, app_cred['access_rules'][0]['id']
            )
            c.get(path, headers=self.headers)

    def test_user_can_list_their_access_rules(self):
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': self.user_id,
            'project_id': self.project_id,
            'secret': uuid.uuid4().hex,
            'access_rules': [
                {
                    'id': uuid.uuid4().hex,
                    'service': uuid.uuid4().hex,
                    'path': uuid.uuid4().hex,
                    'method': uuid.uuid4().hex[16:],
                }
            ],
        }
        PROVIDERS.application_credential_api.create_application_credential(
            app_cred
        )
        with self.test_client() as c:
            r = c.get(
                f'/v3/users/{self.user_id}/access_rules', headers=self.headers
            )
            self.assertEqual(len(r.json['access_rules']), 1)

    def test_user_can_delete_their_access_rules(self):
        access_rule_id = uuid.uuid4().hex
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': self.user_id,
            'project_id': self.project_id,
            'secret': uuid.uuid4().hex,
            'access_rules': [
                {
                    'id': access_rule_id,
                    'service': uuid.uuid4().hex,
                    'path': uuid.uuid4().hex,
                    'method': uuid.uuid4().hex[16:],
                }
            ],
        }
        PROVIDERS.application_credential_api.create_application_credential(
            app_cred
        )
        PROVIDERS.application_credential_api.delete_application_credential(
            app_cred['id']
        )
        with self.test_client() as c:
            path = f'/v3/users/{self.user_id}/access_rules/{access_rule_id}'
            c.delete(path, headers=self.headers)


class _ProjectUsersTests:
    """Users who have project role authorization observe the same behavior."""

    def test_user_cannot_get_access_rules_for_other_users(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        access_rule_id = uuid.uuid4().hex
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': user['id'],
            'project_id': project['id'],
            'secret': uuid.uuid4().hex,
            'access_rules': [
                {
                    'id': access_rule_id,
                    'service': uuid.uuid4().hex,
                    'path': uuid.uuid4().hex,
                    'method': uuid.uuid4().hex[16:],
                }
            ],
        }
        PROVIDERS.application_credential_api.create_application_credential(
            app_cred
        )
        with self.test_client() as c:
            path = '/v3/users/{}/access_rules/{}'.format(
                user['id'], access_rule_id
            )
            c.get(
                path,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_get_own_non_existent_access_rule_not_found(self):
        with self.test_client() as c:
            c.get(
                f'/v3/users/{self.user_id}/access_rules/{uuid.uuid4().hex}',
                headers=self.headers,
                expected_status_code=http.client.NOT_FOUND,
            )

    def test_cannot_get_non_existent_access_rule_other_user_forbidden(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        with self.test_client() as c:
            c.get(
                '/v3/users/{}/access_rules/{}'.format(
                    user['id'], uuid.uuid4().hex
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_list_access_rules_for_other_users(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': user['id'],
            'project_id': project['id'],
            'secret': uuid.uuid4().hex,
            'access_rules': [
                {
                    'id': uuid.uuid4().hex,
                    'service': uuid.uuid4().hex,
                    'path': uuid.uuid4().hex,
                    'method': uuid.uuid4().hex[16:],
                }
            ],
        }
        PROVIDERS.application_credential_api.create_application_credential(
            app_cred
        )

        with self.test_client() as c:
            path = '/v3/users/{}/access_rules'.format(user['id'])
            c.get(
                path,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_access_rules_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )
        access_rule_id = uuid.uuid4().hex
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': user['id'],
            'project_id': project['id'],
            'secret': uuid.uuid4().hex,
            'access_rules': [
                {
                    'id': access_rule_id,
                    'service': uuid.uuid4().hex,
                    'path': uuid.uuid4().hex,
                    'method': uuid.uuid4().hex[16:],
                }
            ],
        }
        PROVIDERS.application_credential_api.create_application_credential(
            app_cred
        )
        PROVIDERS.application_credential_api.delete_application_credential(
            app_cred['id']
        )
        with self.test_client() as c:
            path = '/v3/users/{}/access_rules/{}'.format(
                user['id'], access_rule_id
            )
            c.delete(
                path,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_delete_non_existent_access_rule_other_user_forbidden(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        with self.test_client() as c:
            c.delete(
                '/v3/users/{}/access_rules/{}'.format(
                    user['id'], uuid.uuid4().hex
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class _SystemUserAccessRuleTests:
    """Tests that are common across all system users."""

    def test_user_can_list_access_rules_for_other_users(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': user['id'],
            'project_id': project['id'],
            'secret': uuid.uuid4().hex,
            'access_rules': [
                {
                    'id': uuid.uuid4().hex,
                    'service': uuid.uuid4().hex,
                    'path': uuid.uuid4().hex,
                    'method': uuid.uuid4().hex[16:],
                }
            ],
        }
        PROVIDERS.application_credential_api.create_application_credential(
            app_cred
        )

        with self.test_client() as c:
            r = c.get(
                '/v3/users/{}/access_rules'.format(user['id']),
                headers=self.headers,
            )
            self.assertEqual(1, len(r.json['access_rules']))

    def test_user_cannot_get_non_existent_access_rule_not_found(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        with self.test_client() as c:
            c.get(
                '/v3/users/{}/access_rules/{}'.format(
                    user['id'], uuid.uuid4().hex
                ),
                headers=self.headers,
                expected_status_code=http.client.NOT_FOUND,
            )


class SystemReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserAccessRuleTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_reader)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.reader_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_reader['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_cannot_delete_access_rules_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        access_rule_id = uuid.uuid4().hex
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': user['id'],
            'project_id': project['id'],
            'secret': uuid.uuid4().hex,
            'access_rules': [
                {
                    'id': access_rule_id,
                    'service': uuid.uuid4().hex,
                    'path': uuid.uuid4().hex,
                    'method': uuid.uuid4().hex[16:],
                }
            ],
        }
        PROVIDERS.application_credential_api.create_application_credential(
            app_cred
        )
        PROVIDERS.application_credential_api.delete_application_credential(
            app_cred['id']
        )
        with self.test_client() as c:
            path = '/v3/users/{}/access_rules/{}'.format(
                user['id'], access_rule_id
            )
            c.delete(
                path,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_non_existent_access_rule_forbidden(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        with self.test_client() as c:
            c.delete(
                '/v3/users/{}/access_rules/{}'.format(
                    user['id'], uuid.uuid4().hex
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class SystemMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserAccessRuleTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_member = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_member)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.member_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_member['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_cannot_delete_access_rules_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        access_rule_id = uuid.uuid4().hex
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': user['id'],
            'project_id': project['id'],
            'secret': uuid.uuid4().hex,
            'access_rules': [
                {
                    'id': access_rule_id,
                    'service': uuid.uuid4().hex,
                    'path': uuid.uuid4().hex,
                    'method': uuid.uuid4().hex[16:],
                }
            ],
        }
        PROVIDERS.application_credential_api.create_application_credential(
            app_cred
        )
        PROVIDERS.application_credential_api.delete_application_credential(
            app_cred['id']
        )
        with self.test_client() as c:
            path = '/v3/users/{}/access_rules/{}'.format(
                user['id'], access_rule_id
            )
            c.delete(
                path,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

        with self.test_client() as c:
            path = '/v3/users/{}/access_rules/{}'.format(
                user['id'], access_rule_id
            )
            c.delete(
                path,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_non_existent_access_rule_forbidden(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        with self.test_client() as c:
            c.delete(
                '/v3/users/{}/access_rules/{}'.format(
                    user['id'], uuid.uuid4().hex
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class SystemAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserAccessRuleTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        # Reuse the system administrator account created during
        # ``keystone-manage bootstrap``
        self.user_id = self.bootstrapper.admin_user_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_can_delete_access_rules_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )
        access_rule_id = uuid.uuid4().hex
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': user['id'],
            'project_id': project['id'],
            'secret': uuid.uuid4().hex,
            'access_rules': [
                {
                    'id': access_rule_id,
                    'service': uuid.uuid4().hex,
                    'path': uuid.uuid4().hex,
                    'method': uuid.uuid4().hex[16:],
                }
            ],
        }
        PROVIDERS.application_credential_api.create_application_credential(
            app_cred
        )
        PROVIDERS.application_credential_api.delete_application_credential(
            app_cred['id']
        )

        with self.test_client() as c:
            path = '/v3/users/{}/access_rules/{}'.format(
                user['id'], access_rule_id
            )
            c.delete(path, headers=self.headers)

    def test_user_cannot_delete_non_existent_access_rule_not_found(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        with self.test_client() as c:
            c.delete(
                '/v3/users/{}/access_rules/{}'.format(
                    user['id'], uuid.uuid4().hex
                ),
                headers=self.headers,
                expected_status_code=http.client.NOT_FOUND,
            )


class ProjectReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _UserAccessRuleTests,
    _ProjectUsersTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        project_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(project_reader)['id']
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.project_id = PROVIDERS.resource_api.create_project(
            project['id'], project
        )['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=self.user_id,
            project_id=self.project_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=project_reader['password'],
            project_id=self.project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _UserAccessRuleTests,
    _ProjectUsersTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        project_member = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(project_member)['id']
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.project_id = PROVIDERS.resource_api.create_project(
            project['id'], project
        )['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=self.user_id,
            project_id=self.project_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=project_member['password'],
            project_id=self.project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _UserAccessRuleTests,
    _ProjectUsersTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        # Reuse the system administrator account created during
        # ``keystone-manage bootstrap``
        self.user_id = self.bootstrapper.admin_user_id
        self.project_id = self.bootstrapper.project_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            project_id=self.project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import binascii
import os

from cryptography.exceptions import InvalidKey
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt as Scrypt_kdf

from keystone.common import password_hashers
from keystone import exception


class Scrypt(password_hashers.PasswordHasher):
    """passlib transition class for implementing scrypt password hashing"""

    name: str = "scrypt"
    ident_values: set[str] = {"$scrypt$", "$7$"}

    @staticmethod
    def hash(
        password: bytes,
        salt_size: int = 16,
        n: int = 16,
        r: int = 8,
        p: int = 1,
        **kwargs,
    ) -> str:
        """Generate password hash string with ident and params

        https://docs.python.org/3/library/hashlib.html#hashlib.scrypt

        :param bytes password: Password to be hashed.
        :param int salt_size: Salt size.
        :param int n: CPU/Memory cost factor.
        :param int r: Block size.
        :param int p: Parallel count.

        :returns: String in format `$scrypt$ln=logN,r=R,p=P$salt$checksum`
        """
        salt: bytes = os.urandom(salt_size)
        # Prepare the kdf function
        kdf = Scrypt_kdf(salt=salt, length=32, n=2**n, r=r, p=p)
        # derive - build a digest
        digest = kdf.derive(password)
        # convert digest to string using stripped base64
        digest_str: str = (
            binascii.b2a_base64(digest).rstrip(b"=\n").decode("ascii")
        )
        # apply the same for the salt
        salt_str: str = (
            binascii.b2a_base64(salt).rstrip(b"=\n").decode("ascii")
        )
        return f"$scrypt$ln={n},r={r},p={p}${salt_str}${digest_str}"

    @staticmethod
    def verify(password: bytes, hashed: str) -> bool:
        """Verify hashing password would be equal to the `hashed` value

        :param bytes password: Password to verify
        :param string hashed: Hashed password. Used to extract hashing
            parameters

        :returns: boolean whether hashing password with the same parameters
            would match hashed value
        """
        data: str = hashed
        # split hashed string to extract parameters
        parts: list[str] = data[1:].split("$")
        salt: bytes
        digest: bytes
        n: int
        p: int
        r: int

        if len(parts) == 4:
            ident, params, salt_str, digest_str = parts
            salt = password_hashers.b64s_decode(
                salt_str.replace(".", "+").encode("ascii")
            )
            digest = password_hashers.b64s_decode(
                digest_str.replace(".", "+").encode("ascii")
            )
        else:
            raise exception.PasswordValidationError("malformed password hash")

        for param in params.split(","):
            if param.startswith("ln="):
                n = 2 ** int(param[3:])
            elif param.startswith("p="):
                p = int(param[2:])
            elif param.startswith("r="):
                r = int(param[2:])

        # Prepare the kdf function
        kdf = Scrypt_kdf(salt=salt, length=32, n=n, r=r, p=p)
        # Cryptography raises exception on mismatch
        try:
            kdf.verify(password, digest)
            return True
        except InvalidKey:
            return False



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# NOTE(morgan): This file does not implement any API-specific code. It simply
# supplies shared functions between the "role_inferences" and "roles" (for
# implied roles) APIs. In general, all code for an API should be isolated to
# it's own keystone.api.XXX module and not in the _shared module.

from keystone.common import provider_api
from keystone.server import flask as ks_flask

PROVIDERS = provider_api.ProviderAPIs


def build_prior_role_response_data(prior_role_id, prior_role_name):
    return {
        'id': prior_role_id,
        'links': {'self': ks_flask.base_url(path=f'/roles/{prior_role_id}')},
        'name': prior_role_name,
    }


def build_implied_role_response_data(implied_role):
    return {
        'id': implied_role['id'],
        'links': {
            'self': ks_flask.base_url(
                path='/roles/{}'.format(implied_role['id'])
            )
        },
        'name': implied_role['name'],
    }


def role_inference_response(prior_role_id):
    prior_role = PROVIDERS.role_api.get_role(prior_role_id)
    response = {
        'role_inference': {
            'prior_role': build_prior_role_response_data(
                prior_role_id, prior_role['name']
            )
        }
    }
    return response



# Copyright 2018 SUSE Linux GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystone.auth import plugins as auth_plugins
from keystone.auth.plugins import base
from keystone.common import provider_api
from keystone import exception

PROVIDERS = provider_api.ProviderAPIs
METHOD_NAME = 'application_credential'


class ApplicationCredential(base.AuthMethodHandler):
    def authenticate(self, auth_payload):
        """Authenticate an application."""
        response_data = {}
        app_cred_info = auth_plugins.AppCredInfo.create(
            auth_payload, METHOD_NAME
        )

        try:
            PROVIDERS.application_credential_api.authenticate(
                application_credential_id=app_cred_info.id,
                secret=app_cred_info.secret,
            )
        except AssertionError as e:
            raise exception.Unauthorized(e)
        response_data['user_id'] = app_cred_info.user_id

        return base.AuthHandlerResponse(
            status=True, response_body=None, response_data=response_data
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from keystone.auth import schema
from keystone import exception
from keystone.tests import unit


class TestValidateIssueTokenAuth(unit.BaseTestCase):
    def _expect_failure(self, post_data):
        self.assertRaises(
            exception.SchemaValidationError,
            schema.validate_issue_token_auth,
            post_data,
        )

    def test_auth_not_object_ex(self):
        self._expect_failure('something')

    def test_auth_no_identity_ex(self):
        self._expect_failure({})

    def test_identity_not_object_ex(self):
        self._expect_failure({'identity': 'something'})

    def test_no_methods_ex(self):
        self._expect_failure({'identity': {}})

    def test_methods_not_array_ex(self):
        p = {'identity': {'methods': 'something'}}
        self._expect_failure(p)

    def test_methods_not_array_str_ex(self):
        p = {'identity': {'methods': [{}]}}
        self._expect_failure(p)

    def test_no_auth_plugin_parameters(self):
        # auth plugin (password / token) may not be present.
        post_data = {'identity': {'methods': ['password']}}
        schema.validate_issue_token_auth(post_data)

    def test_password_not_object_ex(self):
        # if password is present, it must be an object.
        p = {'identity': {'methods': ['password'], 'password': 'something'}}
        self._expect_failure(p)

    def test_password_user_not_object_ex(self):
        # if user is present, it must be an object
        p = {
            'identity': {
                'methods': ['password'],
                'password': {'user': 'something'},
            }
        }
        self._expect_failure(p)

    def test_password_user_name_not_string_ex(self):
        # if user name is present, it must be a string
        p = {
            'identity': {
                'methods': ['password'],
                'password': {'user': {'name': 1}},
            }
        }
        self._expect_failure(p)

    def test_password_user_id_not_string_ex(self):
        # if user id is present, it must be a string
        p = {
            'identity': {
                'methods': ['password'],
                'password': {'user': {'id': {}}},
            }
        }
        self._expect_failure(p)

    def test_password_no_user_id_or_name_ex(self):
        # either user id or name must be present.
        p = {'identity': {'methods': ['password'], 'password': {'user': {}}}}
        self._expect_failure(p)

    def test_password_user_password_not_string_ex(self):
        # if user password is present, it must be a string
        p = {
            'identity': {
                'methods': ['password'],
                'password': {'user': {'id': 'something', 'password': {}}},
            }
        }
        self._expect_failure(p)

    def test_password_user_domain_not_object_ex(self):
        # if user domain is present, it must be an object
        p = {
            'identity': {
                'methods': ['password'],
                'password': {
                    'user': {'id': 'something', 'domain': 'something'}
                },
            }
        }
        self._expect_failure(p)

    def test_password_user_domain_no_id_or_name_ex(self):
        # user domain must have id or name.
        p = {
            'identity': {
                'methods': ['password'],
                'password': {'user': {'id': 'something', 'domain': {}}},
            }
        }
        self._expect_failure(p)

    def test_password_user_domain_name_not_string_ex(self):
        # if user domain name is present, it must be a string.
        p = {
            'identity': {
                'methods': ['password'],
                'password': {
                    'user': {'id': 'something', 'domain': {'name': {}}}
                },
            }
        }
        self._expect_failure(p)

    def test_password_user_domain_id_not_string_ex(self):
        # if user domain id is present, it must be a string.
        p = {
            'identity': {
                'methods': ['password'],
                'password': {
                    'user': {'id': 'something', 'domain': {'id': {}}}
                },
            }
        }
        self._expect_failure(p)

    def test_token(self):
        # valid token auth plugin data is supported.
        p = {'identity': {'methods': ['token'], 'token': {'id': 'something'}}}
        schema.validate_issue_token_auth(p)

    def test_token_not_object_ex(self):
        # if token auth plugin data is present, it must be an object.
        p = {'identity': {'methods': ['token'], 'token': ''}}
        self._expect_failure(p)

    def test_token_no_id_ex(self):
        # if token auth plugin data is present, id must be present.
        p = {'identity': {'methods': ['token'], 'token': {}}}
        self._expect_failure(p)

    def test_token_id_not_string_ex(self):
        # if token auth plugin data is present, id must be a string.
        p = {'identity': {'methods': ['token'], 'token': {'id': 123}}}
        self._expect_failure(p)

    def test_scope_not_object_or_string_ex(self):
        p = {'identity': {'methods': []}, 'scope': 1}
        self._expect_failure(p)

    def test_project_not_object_ex(self):
        p = {'identity': {'methods': []}, 'scope': {'project': 'something'}}
        self._expect_failure(p)

    def test_project_name_not_string_ex(self):
        p = {'identity': {'methods': []}, 'scope': {'project': {'name': {}}}}
        self._expect_failure(p)

    def test_project_id_not_string_ex(self):
        p = {'identity': {'methods': []}, 'scope': {'project': {'id': {}}}}
        self._expect_failure(p)

    def test_project_no_id_or_name_ex(self):
        p = {'identity': {'methods': []}, 'scope': {'project': {}}}
        self._expect_failure(p)

    def test_project_domain_not_object_ex(self):
        p = {
            'identity': {'methods': []},
            'scope': {'project': {'id': 'something', 'domain': 'something'}},
        }
        self._expect_failure(p)

    def test_project_domain_name_not_string_ex(self):
        p = {
            'identity': {'methods': []},
            'scope': {'project': {'id': 'something', 'domain': {'name': {}}}},
        }
        self._expect_failure(p)

    def test_project_domain_id_not_string_ex(self):
        p = {
            'identity': {'methods': []},
            'scope': {'project': {'id': 'something', 'domain': {'id': {}}}},
        }
        self._expect_failure(p)

    def test_project_domain_no_id_or_name_ex(self):
        p = {
            'identity': {'methods': []},
            'scope': {'project': {'id': 'something', 'domain': {}}},
        }
        self._expect_failure(p)

    def test_domain_not_object_ex(self):
        p = {'identity': {'methods': []}, 'scope': {'domain': 'something'}}
        self._expect_failure(p)

    def test_domain_id_not_string_ex(self):
        p = {'identity': {'methods': []}, 'scope': {'domain': {'id': {}}}}
        self._expect_failure(p)

    def test_domain_name_not_string_ex(self):
        p = {'identity': {'methods': []}, 'scope': {'domain': {'name': {}}}}
        self._expect_failure(p)

    def test_domain_no_id_or_name_ex(self):
        p = {'identity': {'methods': []}, 'scope': {'domain': {}}}
        self._expect_failure(p)

    def test_trust_not_object_ex(self):
        p = {
            'identity': {'methods': []},
            'scope': {'OS-TRUST:trust': 'something'},
        }
        self._expect_failure(p)

    def test_unscoped(self):
        post_data = {
            'identity': {
                'methods': ['password'],
                'password': {
                    'user': {
                        'name': 'admin',
                        'domain': {'name': 'Default'},
                        'password': 'devstacker',
                    }
                },
            }
        }
        schema.validate_issue_token_auth(post_data)

    def test_user_domain_id(self):
        post_data = {
            'identity': {
                'methods': ['password'],
                'password': {
                    'user': {
                        'name': 'admin',
                        'domain': {'id': 'default'},
                        'password': 'devstacker',
                    }
                },
            }
        }
        schema.validate_issue_token_auth(post_data)

    def test_two_methods(self):
        post_data = {
            'identity': {
                'methods': ['password', 'mapped'],
                'password': {
                    'user': {
                        'name': 'admin',
                        'domain': {'name': 'Default'},
                        'password': 'devstacker',
                    }
                },
            }
        }
        schema.validate_issue_token_auth(post_data)

    def test_project_scoped(self):
        post_data = {
            'identity': {
                'methods': ['password'],
                'password': {
                    'user': {
                        'name': 'admin',
                        'domain': {'name': 'Default'},
                        'password': 'devstacker',
                    }
                },
            },
            'scope': {
                'project': {'name': 'demo', 'domain': {'name': 'Default'}}
            },
        }
        schema.validate_issue_token_auth(post_data)

    def test_domain_scoped(self):
        post_data = {
            'identity': {
                'methods': ['password'],
                'password': {
                    'user': {
                        'name': 'admin',
                        'domain': {'name': 'Default'},
                        'password': 'devstacker',
                    }
                },
            },
            'scope': {'domain': {'name': 'Default'}},
        }
        schema.validate_issue_token_auth(post_data)

    def test_explicit_unscoped(self):
        post_data = {
            'identity': {
                'methods': ['password'],
                'password': {
                    'user': {
                        'name': 'admin',
                        'domain': {'name': 'Default'},
                        'password': 'devstacker',
                    }
                },
            },
            'scope': 'unscoped',
        }
        schema.validate_issue_token_auth(post_data)

    def test_additional_properties(self):
        # Everything can have extra properties and they're ignored.
        p = {
            'identity': {
                'methods': ['password'],
                'password': {
                    'user': {
                        'id': 'whatever',
                        'extra4': 'whatever4',
                        'domain': {'id': 'whatever', 'extra5': 'whatever5'},
                    },
                    'extra3': 'whatever3',
                },
                'token': {'id': 'something', 'extra9': 'whatever9'},
                'extra4': 'whatever4',
            },
            'scope': {
                'project': {
                    'id': 'something',
                    'domain': {'id': 'something', 'extra8': 'whatever8'},
                    'extra7': 'whatever7',
                },
                'domain': {'id': 'something', 'extra9': 'whatever9'},
                'extra6': 'whatever6',
            },
            'extra2': 'whatever2',
        }
        schema.validate_issue_token_auth(p)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import http.client
import uuid

from oslo_serialization import jsonutils
from oslo_utils import timeutils

from keystone.common.policies import base as base_policy
from keystone.common import provider_api
import keystone.conf
from keystone.tests.common import auth as common_auth
from keystone.tests import unit
from keystone.tests.unit import base_classes
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import temporaryfile

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class _TestAppCredBase(base_classes.TestCaseWithBootstrap):
    """Base class for application credential tests."""

    def _new_app_cred_data(
        self,
        user_id=None,
        project_id=None,
        name=None,
        expires=None,
        system=None,
    ):
        if not user_id:
            user_id = self.app_cred_user_id
        if not name:
            name = uuid.uuid4().hex
        if not expires:
            expires = timeutils.utcnow() + datetime.timedelta(days=365)
        if not system:
            system = uuid.uuid4().hex
        if not project_id:
            project_id = self.app_cred_project_id
        app_cred_data = {
            'id': uuid.uuid4().hex,
            'name': name,
            'description': uuid.uuid4().hex,
            'user_id': user_id,
            'project_id': project_id,
            'system': system,
            'expires_at': expires,
            'roles': [{'id': self.bootstrapper.member_role_id}],
            'secret': uuid.uuid4().hex,
            'unrestricted': False,
        }
        return app_cred_data

    def setUp(self):
        super().setUp()

        # create a user and project for app cred testing
        new_user_ref = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        app_cred_user_ref = PROVIDERS.identity_api.create_user(new_user_ref)
        self.app_cred_user_id = app_cred_user_ref['id']
        self.app_cred_user_password = new_user_ref['password']
        app_cred_project_ref = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )
        self.app_cred_project_id = app_cred_project_ref['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=self.app_cred_user_id,
            project_id=self.app_cred_project_id,
        )

    def _create_application_credential(self):
        app_cred = self._new_app_cred_data()
        return (
            PROVIDERS.application_credential_api.create_application_credential(
                app_cred
            )
        )

    def _override_policy(self):
        # TODO(gyee): Remove this once the deprecated policies in
        # keystone.common.policies.application_credential have been removed.
        # This is only here to make sure we test the new policies instead of
        # the deprecated ones. Oslo.policy will OR deprecated policies with
        # new policies to maintain compatibility and give operators a chance to
        # update permissions or update policies without breaking users.
        # This will cause these specific tests to fail since we're trying to
        # correct this broken behavior with better scope checking.
        with open(self.policy_file_name, 'w') as f:
            overridden_policies = {
                'identity:get_application_credential': (
                    base_policy.RULE_SYSTEM_READER_OR_OWNER
                ),
                'identity:list_application_credentials': (
                    base_policy.RULE_SYSTEM_READER_OR_OWNER
                ),
                'identity:create_application_credential': (
                    base_policy.RULE_OWNER
                ),
                'identity:delete_application_credential': (
                    base_policy.RULE_SYSTEM_ADMIN_OR_OWNER
                ),
            }
            f.write(jsonutils.dumps(overridden_policies))


class _DomainAndProjectUserTests:
    """Domain and project user tests.

    Domain and project users should not be able to manage application
    credentials other then their own.
    """

    def test_user_cannot_list_application_credentials(self):
        # create a couple of application credentials
        self._create_application_credential()
        self._create_application_credential()

        with self.test_client() as c:
            c.get(
                f'/v3/users/{self.app_cred_user_id}/application_credentials',
                expected_status_code=http.client.FORBIDDEN,
                headers=self.headers,
            )

    def test_user_cannot_get_application_credential(self):
        app_cred = self._create_application_credential()

        with self.test_client() as c:
            c.get(
                '/v3/users/{}/application_credentials/{}'.format(
                    self.app_cred_user_id, app_cred['id']
                ),
                expected_status_code=http.client.FORBIDDEN,
                headers=self.headers,
            )

    def test_user_cannot_lookup_application_credential(self):
        app_cred = self._create_application_credential()

        with self.test_client() as c:
            c.get(
                '/v3/users/{}/application_credentials?name={}'.format(
                    self.app_cred_user_id, app_cred['name']
                ),
                expected_status_code=http.client.FORBIDDEN,
                headers=self.headers,
            )

    def test_user_cannot_delete_application_credential(self):
        app_cred = self._create_application_credential()

        with self.test_client() as c:
            c.delete(
                '/v3/users/{}/application_credentials/{}'.format(
                    self.app_cred_user_id, app_cred['id']
                ),
                expected_status_code=http.client.FORBIDDEN,
                headers=self.headers,
            )

    def test_user_cannot_lookup_non_existent_application_credential(self):
        with self.test_client() as c:
            c.get(
                f'/v3/users/{self.app_cred_user_id}/application_credentials?name={uuid.uuid4().hex}',
                expected_status_code=http.client.FORBIDDEN,
                headers=self.headers,
            )

    def test_user_cannot_create_app_credential_for_another_user(self):
        # create another user
        another_user = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        another_user_id = PROVIDERS.identity_api.create_user(another_user)[
            'id'
        ]

        app_cred_body = {
            'application_credential': unit.new_application_credential_ref(
                roles=[{'id': self.bootstrapper.member_role_id}]
            )
        }

        with self.test_client() as c:
            c.post(
                f'/v3/users/{another_user_id}/application_credentials',
                json=app_cred_body,
                expected_status_code=http.client.FORBIDDEN,
                headers=self.headers,
            )


class _SystemUserAndOwnerTests:
    """Common default functionality for all system users and owner."""

    def test_user_can_list_application_credentials(self):
        # create a couple of application credentials
        self._create_application_credential()
        self._create_application_credential()

        with self.test_client() as c:
            r = c.get(
                f'/v3/users/{self.app_cred_user_id}/application_credentials',
                headers=self.headers,
            )
            self.assertEqual(2, len(r.json['application_credentials']))

    def test_user_can_get_application_credential(self):
        app_cred = self._create_application_credential()

        with self.test_client() as c:
            r = c.get(
                '/v3/users/{}/application_credentials/{}'.format(
                    self.app_cred_user_id, app_cred['id']
                ),
                headers=self.headers,
            )
            actual_app_cred = r.json['application_credential']
            self.assertEqual(app_cred['id'], actual_app_cred['id'])

    def test_user_can_lookup_application_credential(self):
        app_cred = self._create_application_credential()

        with self.test_client() as c:
            r = c.get(
                '/v3/users/{}/application_credentials?name={}'.format(
                    self.app_cred_user_id, app_cred['name']
                ),
                headers=self.headers,
            )
            self.assertEqual(1, len(r.json['application_credentials']))
            actual_app_cred = r.json['application_credentials'][0]
            self.assertEqual(app_cred['id'], actual_app_cred['id'])

    def _test_delete_application_credential(
        self, expected_status_code=http.client.NO_CONTENT
    ):
        app_cred = self._create_application_credential()

        with self.test_client() as c:
            c.delete(
                '/v3/users/{}/application_credentials/{}'.format(
                    self.app_cred_user_id, app_cred['id']
                ),
                expected_status_code=expected_status_code,
                headers=self.headers,
            )

    def test_user_cannot_create_app_credential_for_another_user(self):
        # create another user
        another_user = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        another_user_id = PROVIDERS.identity_api.create_user(another_user)[
            'id'
        ]

        app_cred_body = {
            'application_credential': unit.new_application_credential_ref(
                roles=[{'id': self.bootstrapper.member_role_id}]
            )
        }

        with self.test_client() as c:
            c.post(
                f'/v3/users/{another_user_id}/application_credentials',
                json=app_cred_body,
                expected_status_code=http.client.FORBIDDEN,
                headers=self.headers,
            )


class SystemReaderTests(
    _TestAppCredBase, common_auth.AuthTestMixin, _SystemUserAndOwnerTests
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_reader)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.reader_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_reader['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_system_reader_cannot_delete_application_credential_for_user(self):
        self._test_delete_application_credential(
            expected_status_code=http.client.FORBIDDEN
        )


class SystemMemberTests(
    _TestAppCredBase, common_auth.AuthTestMixin, _SystemUserAndOwnerTests
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_member = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_member)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.member_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_member['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_system_reader_cannot_delete_application_credential_for_user(self):
        self._test_delete_application_credential(
            expected_status_code=http.client.FORBIDDEN
        )


class SystemAdminTests(
    _TestAppCredBase, common_auth.AuthTestMixin, _SystemUserAndOwnerTests
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        self.user_id = self.bootstrapper.admin_user_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_system_admin_can_delete_application_credential_for_user(self):
        self._test_delete_application_credential()


class OwnerTests(
    _TestAppCredBase, common_auth.AuthTestMixin, _SystemUserAndOwnerTests
):
    def setUp(self):
        super().setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )

        self._override_policy()
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        # in this case app_cred_user_id and user_id are the same since we
        # are testing the owner
        self.user_id = self.app_cred_user_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.app_cred_user_password,
            project_id=self.app_cred_project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_create_application_credential_by_owner(self):
        app_cred_body = {
            'application_credential': unit.new_application_credential_ref()
        }

        with self.test_client() as c:
            c.post(
                f'/v3/users/{self.user_id}/application_credentials',
                json=app_cred_body,
                expected_status_code=http.client.CREATED,
                headers=self.headers,
            )

    def test_owner_can_delete_application_credential(self):
        self._test_delete_application_credential()

    def test_user_cannot_lookup_application_credential_for_another_user(self):
        # create another user
        another_user = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        another_user_id = PROVIDERS.identity_api.create_user(another_user)[
            'id'
        ]

        auth = self.build_authentication_request(
            user_id=another_user_id, password=another_user['password']
        )

        # authenticate for a token as a completely different user with
        # completely different authorization
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            another_user_token = r.headers['X-Subject-Token']

        # create an application credential as the self.user_id user on a
        # project that the user above doesn't have any authorization on
        app_cred = self._create_application_credential()

        # attempt to lookup the application credential as another user
        with self.test_client() as c:
            c.get(
                '/v3/users/{}/application_credentials/{}'.format(
                    another_user_id, app_cred['id']
                ),
                expected_status_code=http.client.FORBIDDEN,
                headers={'X-Auth-Token': another_user_token},
            )

    def test_user_cannot_delete_application_credential_for_another_user(self):
        # create another user
        another_user = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        another_user_id = PROVIDERS.identity_api.create_user(another_user)[
            'id'
        ]

        auth = self.build_authentication_request(
            user_id=another_user_id, password=another_user['password']
        )

        # authenticate for a token as a completely different user with
        # completely different authorization
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            another_user_token = r.headers['X-Subject-Token']

        # create an application credential as the self.user_id user on a
        # project that the user above doesn't have any authorization on
        app_cred = self._create_application_credential()

        # attempt to delete the application credential as another user
        with self.test_client() as c:
            c.delete(
                '/v3/users/{}/application_credentials/{}'.format(
                    another_user_id, app_cred['id']
                ),
                expected_status_code=http.client.FORBIDDEN,
                headers={'X-Auth-Token': another_user_token},
            )


class DomainAdminTests(
    _TestAppCredBase, common_auth.AuthTestMixin, _DomainAndProjectUserTests
):
    def setUp(self):
        super().setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )

        self._override_policy()
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain_admin = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(domain_admin)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.admin_role_id,
            user_id=self.user_id,
            domain_id=CONF.identity.default_domain_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_admin['password'],
            domain_id=CONF.identity.default_domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class DomainReaderTests(
    _TestAppCredBase, common_auth.AuthTestMixin, _DomainAndProjectUserTests
):
    def setUp(self):
        super().setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )

        self._override_policy()
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain_admin = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(domain_admin)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=self.user_id,
            domain_id=CONF.identity.default_domain_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_admin['password'],
            domain_id=CONF.identity.default_domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class DomainMemberTests(
    _TestAppCredBase, common_auth.AuthTestMixin, _DomainAndProjectUserTests
):
    def setUp(self):
        super().setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )

        self._override_policy()
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain_admin = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(domain_admin)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=self.user_id,
            domain_id=CONF.identity.default_domain_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_admin['password'],
            domain_id=CONF.identity.default_domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectAdminTests(
    _TestAppCredBase, common_auth.AuthTestMixin, _DomainAndProjectUserTests
):
    def setUp(self):
        super().setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )

        self._override_policy()
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        project_admin = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(project_admin)['id']
        # even project admin of project where the app credential
        # is intended for cannot perform app credential operations
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.admin_role_id,
            user_id=self.user_id,
            project_id=self.app_cred_project_id,
        )
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=project_admin['password'],
            project_id=self.app_cred_project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectReaderTests(
    _TestAppCredBase, common_auth.AuthTestMixin, _DomainAndProjectUserTests
):
    def setUp(self):
        super().setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )

        self._override_policy()
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        project_admin = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(project_admin)['id']
        # even project admin of project where the app credential
        # is intended for cannot perform app credential operations
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=self.user_id,
            project_id=self.app_cred_project_id,
        )
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=project_admin['password'],
            project_id=self.app_cred_project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectMemberTests(
    _TestAppCredBase, common_auth.AuthTestMixin, _DomainAndProjectUserTests
):
    def setUp(self):
        super().setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )

        self._override_policy()
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        project_admin = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(project_admin)['id']
        # even project admin of project where the app credential
        # is intended for cannot perform app credential operations
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=self.user_id,
            project_id=self.app_cred_project_id,
        )
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=project_admin['password'],
            project_id=self.app_cred_project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import http.client
import uuid

from keystone.common import provider_api
import keystone.conf
from keystone.tests.common import auth as common_auth
from keystone.tests import unit
from keystone.tests.unit import base_classes
from keystone.tests.unit import ksfixtures

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class _SystemUserRoleTests:
    """Common default functionality for all system users."""

    def test_user_can_list_roles(self):
        PROVIDERS.role_api.create_role(uuid.uuid4().hex, unit.new_role_ref())

        with self.test_client() as c:
            r = c.get('/v3/roles', headers=self.headers)
            # With bootstrap setup and the role we just created, there should
            # be five roles present in the deployment. Bootstrap creates
            # ``service``, ``admin``, ``member``, and ``reader``.
            self.assertEqual(5, len(r.json['roles']))

    def test_user_can_get_a_role(self):
        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref()
        )

        with self.test_client() as c:
            r = c.get('/v3/roles/{}'.format(role['id']), headers=self.headers)
            self.assertEqual(role['id'], r.json['role']['id'])


class _SystemReaderAndMemberRoleTests:
    """Common default functionality for system readers and system members."""

    def test_user_cannot_create_roles(self):
        create = {'role': unit.new_role_ref()}

        with self.test_client() as c:
            c.post(
                '/v3/roles',
                json=create,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_roles(self):
        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref()
        )

        update = {'role': {'description': uuid.uuid4().hex}}

        with self.test_client() as c:
            c.patch(
                '/v3/roles/{}'.format(role['id']),
                json=update,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_roles(self):
        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref()
        )

        with self.test_client() as c:
            c.delete(
                '/v3/roles/{}'.format(role['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class _DomainAndProjectUserRoleTests:
    """Common functionality for all domain and project users."""

    def test_user_cannot_list_roles(self):
        PROVIDERS.role_api.create_role(uuid.uuid4().hex, unit.new_role_ref())

        with self.test_client() as c:
            c.get(
                '/v3/roles',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_get_a_role(self):
        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref()
        )

        with self.test_client() as c:
            c.get(
                '/v3/roles/{}'.format(role['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_create_roles(self):
        create = {'role': unit.new_role_ref()}

        with self.test_client() as c:
            c.post(
                '/v3/roles',
                json=create,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_roles(self):
        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref()
        )

        update = {'role': {'description': uuid.uuid4().hex}}

        with self.test_client() as c:
            c.patch(
                '/v3/roles/{}'.format(role['id']),
                json=update,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_roles(self):
        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref()
        )

        with self.test_client() as c:
            c.delete(
                '/v3/roles/{}'.format(role['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class SystemReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserRoleTests,
    _SystemReaderAndMemberRoleTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_reader)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.reader_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_reader['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserRoleTests,
    _SystemReaderAndMemberRoleTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_member = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_member)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.member_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_member['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserRoleTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        # Reuse the system administrator account created during
        # ``keystone-manage bootstrap``
        self.user_id = self.bootstrapper.admin_user_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_can_create_roles(self):
        create = {'role': unit.new_role_ref()}

        with self.test_client() as c:
            c.post('/v3/roles', json=create, headers=self.headers)

    def test_user_can_update_roles(self):
        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref()
        )

        update = {'role': {'description': uuid.uuid4().hex}}

        with self.test_client() as c:
            c.patch(
                '/v3/roles/{}'.format(role['id']),
                json=update,
                headers=self.headers,
            )

    def test_user_can_delete_roles(self):
        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref()
        )

        with self.test_client() as c:
            c.delete('/v3/roles/{}'.format(role['id']), headers=self.headers)


class DomainUserTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _DomainAndProjectUserRoleTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_admin = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(domain_admin)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.admin_role_id,
            user_id=self.user_id,
            domain_id=self.domain_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_admin['password'],
            domain_id=self.domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectUserTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _DomainAndProjectUserRoleTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        self.user_id = self.bootstrapper.admin_user_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            project_id=self.bootstrapper.project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectUserTestsWithoutEnforceScope(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _DomainAndProjectUserRoleTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))

        # Explicityly set enforce_scope to False to make sure we maintain
        # backwards compatibility with project users.
        self.config_fixture.config(group='oslo_policy', enforce_scope=False)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        user = unit.new_user_ref(domain_id=domain['id'])
        self.user_id = PROVIDERS.identity_api.create_user(user)['id']

        self.project_id = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=domain['id'])
        )['id']

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=self.user_id,
            project_id=self.project_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=user['password'],
            project_id=self.project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}



# Copyright 2013 OpenStack Foundation
# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from alembic import command as alembic_api
from alembic import config as alembic_config
from alembic import migration as alembic_migration
from alembic import script as alembic_script
from oslo_db import exception as db_exception
from oslo_log import log as logging
from oslo_utils import fileutils

from keystone.common import sql
import keystone.conf

CONF = keystone.conf.CONF
LOG = logging.getLogger(__name__)

ALEMBIC_INIT_VERSION = '27e647c0fad4'

EXPAND_BRANCH = 'expand'
DATA_MIGRATION_BRANCH = 'data_migration'
CONTRACT_BRANCH = 'contract'

RELEASES = ('yoga', 'bobcat', '2024.01')
MILESTONES = (
    'yoga',
    # Do not add the milestone until the end of the release
)
CURRENT_RELEASE = RELEASES[-1]
MIGRATION_BRANCHES = (EXPAND_BRANCH, CONTRACT_BRANCH)
VERSIONS_PATH = os.path.join(
    os.path.dirname(sql.__file__), 'migrations', 'versions'
)


def get_version_branch_path(release=None, branch=None):
    """Get the path to a version branch."""
    version_path = VERSIONS_PATH
    if branch and release:
        return os.path.join(version_path, release, branch)
    return version_path


def check_bootstrap_new_branch(branch, version_path, addn_kwargs):
    """Bootstrap a new migration branch if it does not exist."""
    addn_kwargs['version_path'] = version_path
    addn_kwargs['head'] = f'{branch}@head'
    if not os.path.exists(version_path):
        # Bootstrap initial directory structure
        fileutils.ensure_tree(version_path, mode=0o755)


def _find_alembic_conf():
    """Get the project's alembic configuration.

    :returns: An instance of ``alembic.config.Config``
    """
    path = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), 'alembic.ini'
    )

    config = alembic_config.Config(os.path.abspath(path))

    config.set_main_option('sqlalchemy.url', CONF.database.connection)

    # we don't want to use the logger configuration from the file, which is
    # only really intended for the CLI
    # https://stackoverflow.com/a/42691781/613428
    config.attributes['configure_logger'] = False

    # we want to scan all the versioned subdirectories
    version_paths = [VERSIONS_PATH]
    for release in RELEASES:
        for branch in MIGRATION_BRANCHES:
            version_path = os.path.join(VERSIONS_PATH, release, branch)
            version_paths.append(version_path)
    config.set_main_option('version_locations', ' '.join(version_paths))

    return config


def get_alembic_config():
    return _find_alembic_conf()


def _get_current_heads(engine, config):
    script = alembic_script.ScriptDirectory.from_config(config)

    with engine.connect() as conn:
        context = alembic_migration.MigrationContext.configure(conn)
        heads = context.get_current_heads()

    heads_map = {}

    for head in heads:
        if CONTRACT_BRANCH in script.get_revision(head).branch_labels:
            heads_map[CONTRACT_BRANCH] = head
        else:
            heads_map[EXPAND_BRANCH] = head

    return heads_map


def get_current_heads():
    """Get the current head of each the expand and contract branches."""
    config = _find_alembic_conf()

    with sql.session_for_read() as session:
        engine = session.get_bind()

    # discard the URL encoded in alembic.ini in favour of the URL
    # configured for the engine by the database fixtures, casting from
    # 'sqlalchemy.engine.url.URL' to str in the process. This returns a
    # RFC-1738 quoted URL, which means that a password like "foo@" will be
    # turned into "foo%40". This in turns causes a problem for
    # set_main_option() because that uses ConfigParser.set, which (by
    # design) uses *python* interpolation to write the string out ... where
    # "%" is the special python interpolation character! Avoid this
    # mismatch by quoting all %'s for the set below.
    engine_url = engine.url.render_as_string(hide_password=False).replace(
        '%', '%%'
    )
    config.set_main_option('sqlalchemy.url', engine_url)

    heads = _get_current_heads(engine, config)

    return heads


def _is_database_under_alembic_control(engine):
    with engine.connect() as conn:
        context = alembic_migration.MigrationContext.configure(conn)
        return bool(context.get_current_heads())


def _upgrade_alembic(engine, config, branch):
    revision = 'heads'
    if branch:
        revision = f'{branch}@head'

    # re-use the connection rather than creating a new one
    with engine.begin() as connection:
        config.attributes['connection'] = connection
        alembic_api.upgrade(config, revision)


def get_db_version(branch=EXPAND_BRANCH, *, engine=None):
    config = _find_alembic_conf()

    if engine is None:
        with sql.session_for_read() as session:
            engine = session.get_bind()

    # discard the URL encoded in alembic.ini in favour of the URL
    # configured for the engine by the database fixtures, casting from
    # 'sqlalchemy.engine.url.URL' to str in the process. This returns a
    # RFC-1738 quoted URL, which means that a password like "foo@" will be
    # turned into "foo%40". This in turns causes a problem for
    # set_main_option() because that uses ConfigParser.set, which (by
    # design) uses *python* interpolation to write the string out ... where
    # "%" is the special python interpolation character! Avoid this
    # mismatch by quoting all %'s for the set below.
    engine_url = engine.url.render_as_string(hide_password=False).replace(
        '%', '%%'
    )
    config.set_main_option('sqlalchemy.url', engine_url)

    # we use '.get' since the particular branch might not have been created
    alembic_version = _get_current_heads(engine, config).get(branch)

    return alembic_version


def _db_sync(branch=None, *, engine=None):
    config = _find_alembic_conf()

    if engine is None:
        with sql.session_for_write() as session:
            engine = session.get_bind()

    # discard the URL encoded in alembic.ini in favour of the URL
    # configured for the engine by the database fixtures, casting from
    # 'sqlalchemy.engine.url.URL' to str in the process. This returns a
    # RFC-1738 quoted URL, which means that a password like "foo@" will be
    # turned into "foo%40". This in turns causes a problem for
    # set_main_option() because that uses ConfigParser.set, which (by
    # design) uses *python* interpolation to write the string out ... where
    # "%" is the special python interpolation character! Avoid this
    # mismatch by quoting all %'s for the set below.
    engine_url = engine.url.render_as_string(hide_password=False).replace(
        '%', '%%'
    )
    config.set_main_option('sqlalchemy.url', engine_url)

    _upgrade_alembic(engine, config, branch)


def _validate_upgrade_order(branch, *, engine=None):
    """Validate the upgrade order of the migration branches.

    This is run before allowing the db_sync command to execute. Ensure the
    expand steps have been run before the contract steps.

    :param branch: The name of the branch that the user is trying to
        upgrade.
    """
    if branch == EXPAND_BRANCH:
        return

    if branch == DATA_MIGRATION_BRANCH:
        # this is a no-op in alembic land
        return

    config = _find_alembic_conf()

    if engine is None:
        with sql.session_for_read() as session:
            engine = session.get_bind()

    script = alembic_script.ScriptDirectory.from_config(config)
    expand_head = None
    for head in script.get_heads():
        if EXPAND_BRANCH in script.get_revision(head).branch_labels:
            expand_head = head
            break

    with engine.connect() as conn:
        context = alembic_migration.MigrationContext.configure(conn)
        current_heads = context.get_current_heads()

    if expand_head not in current_heads:
        raise db_exception.DBMigrationError(
            'You are attempting to upgrade contract ahead of expand. '
            'Please refer to '
            'https://docs.openstack.org/keystone/latest/admin/'
            'identity-upgrading.html '
            'to see the proper steps for rolling upgrades.'
        )


def expand_schema(engine=None):
    """Expand the database schema ahead of data migration.

    This is run manually by the keystone-manage command before the first
    keystone node is migrated to the latest release.
    """
    _validate_upgrade_order(EXPAND_BRANCH, engine=engine)
    _db_sync(EXPAND_BRANCH, engine=engine)


def migrate_data(engine=None):
    """Migrate data to match the new schema.

    This is run manually by the keystone-manage command once the keystone
    schema has been expanded for the new release.
    """
    print(
        'Data migrations are no longer supported with alembic. '
        'This is now a no-op.'
    )


def contract_schema(engine=None):
    """Contract the database.

    This is run manually by the keystone-manage command once the keystone
    nodes have been upgraded to the latest release and will remove any old
    tables/columns that are no longer required.
    """
    _validate_upgrade_order(CONTRACT_BRANCH, engine=engine)
    _db_sync(CONTRACT_BRANCH, engine=engine)


def offline_sync_database_to_version(version=None, *, engine=None):
    """Perform and off-line sync of the database.

    Migrate the database up to the latest version, doing the equivalent of
    the cycle of --expand, --migrate and --contract, for when an offline
    upgrade is being performed.

    If a version is specified then only migrate the database up to that
    version. Downgrading is not supported. If version is specified, then only
    the main database migration is carried out - and the expand, migration and
    contract phases will NOT be run.
    """
    if version:
        raise Exception('Specifying a version is no longer supported')

    _db_sync(engine=engine)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# NOTE(morgan): This module contains json_home partial functions for
# what were called "extensions" before. As keystone does not have extensions
# any longer, once Keystone is converted to flask fully, there should be no
# reason to add more elements to this module.

import functools

from keystone.common import json_home

# OS-EC2 "extension"
os_ec2_resource_rel_func = functools.partial(
    json_home.build_v3_extension_resource_relation,
    extension_name='OS-EC2',
    extension_version='1.0',
)

# s3token "extension"
s3_token_resource_rel_func = functools.partial(
    json_home.build_v3_extension_resource_relation,
    extension_name='s3tokens',
    extension_version='1.0',
)

# OS-EP-FILTER "extension"
os_ep_filter_resource_rel_func = functools.partial(
    json_home.build_v3_extension_resource_relation,
    extension_name='OS-EP-FILTER',
    extension_version='1.0',
)
os_ep_filter_parameter_rel_func = functools.partial(
    json_home.build_v3_extension_parameter_relation,
    extension_name='OS-EP-FILTER',
    extension_version='1.0',
)

# OS-OAUTH1 "extension"
os_oauth1_resource_rel_func = functools.partial(
    json_home.build_v3_extension_resource_relation,
    extension_name='OS-OAUTH1',
    extension_version='1.0',
)
os_oauth1_parameter_rel_func = functools.partial(
    json_home.build_v3_extension_parameter_relation,
    extension_name='OS-OAUTH1',
    extension_version='1.0',
)

# OS-OAUTH2 "extension"
os_oauth2_resource_rel_func = functools.partial(
    json_home.build_v3_extension_resource_relation,
    extension_name='OS-OAUTH2',
    extension_version='1.0',
)
os_oauth2_parameter_rel_func = functools.partial(
    json_home.build_v3_extension_parameter_relation,
    extension_name='OS-OAUTH2',
    extension_version='1.0',
)

# OS-REVOKE "extension"
os_revoke_resource_rel_func = functools.partial(
    json_home.build_v3_extension_resource_relation,
    extension_name='OS-REVOKE',
    extension_version='1.0',
)

# OS-SIMPLE-CERT "extension"
os_simple_cert_resource_rel_func = functools.partial(
    json_home.build_v3_extension_resource_relation,
    extension_name='OS-SIMPLE-CERT',
    extension_version='1.0',
)

# OS-TRUST "extension"
os_trust_resource_rel_func = functools.partial(
    json_home.build_v3_extension_resource_relation,
    extension_name='OS-TRUST',
    extension_version='1.0',
)
os_trust_parameter_rel_func = functools.partial(
    json_home.build_v3_extension_parameter_relation,
    extension_name='OS-TRUST',
    extension_version='1.0',
)

# OS-ENDPOINT-POLICY "extension"
os_endpoint_policy_resource_rel_func = functools.partial(
    json_home.build_v3_extension_resource_relation,
    extension_name='OS-ENDPOINT-POLICY',
    extension_version='1.0',
)

# OS-FEDERATION "extension"
os_federation_resource_rel_func = functools.partial(
    json_home.build_v3_extension_resource_relation,
    extension_name='OS-FEDERATION',
    extension_version='1.0',
)
os_federation_parameter_rel_func = functools.partial(
    json_home.build_v3_extension_parameter_relation,
    extension_name='OS-FEDERATION',
    extension_version='1.0',
)

# OS-INHERIT "extension"
os_inherit_resource_rel_func = functools.partial(
    json_home.build_v3_extension_resource_relation,
    extension_name='OS-INHERIT',
    extension_version='1.0',
)

# OS-PKI (revoked) "extension"
os_pki_resource_rel_func = functools.partial(
    json_home.build_v3_extension_resource_relation,
    extension_name='OS-PKI',
    extension_version='1.0',
)



# Copyright 2012 OpenStack Foundation
# Copyright 2012 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sqlalchemy
from sqlalchemy.sql import true

from keystone.catalog.backends import base
from keystone.common import driver_hints
from keystone.common import sql
from keystone.common import utils
import keystone.conf
from keystone import exception
from keystone.i18n import _

CONF = keystone.conf.CONF


class Region(sql.ModelBase, sql.ModelDictMixinWithExtras):
    __tablename__ = 'region'
    attributes = ['id', 'description', 'parent_region_id']
    id = sql.Column(sql.String(255), primary_key=True)
    description = sql.Column(sql.String(255), nullable=False)
    # NOTE(jaypipes): Right now, using an adjacency list model for
    #                 storing the hierarchy of regions is fine, since
    #                 the API does not support any kind of querying for
    #                 more complex hierarchical queries such as "get me only
    #                 the regions that are subchildren of this region", etc.
    #                 If, in the future, such queries are needed, then it
    #                 would be possible to add in columns to this model for
    #                 "left" and "right" and provide support for a nested set
    #                 model.
    parent_region_id = sql.Column(sql.String(255), nullable=True)
    extra = sql.Column(sql.JsonBlob())
    endpoints = sqlalchemy.orm.relationship("Endpoint", backref="region")


class Service(sql.ModelBase, sql.ModelDictMixinWithExtras):
    __tablename__ = 'service'
    attributes = ['id', 'type', 'enabled']
    id = sql.Column(sql.String(64), primary_key=True)
    type = sql.Column(sql.String(255))
    enabled = sql.Column(
        sql.Boolean,
        nullable=False,
        default=True,
        server_default=sqlalchemy.sql.expression.true(),
    )
    extra = sql.Column(sql.JsonBlob())
    endpoints = sqlalchemy.orm.relationship("Endpoint", backref="service")


class Endpoint(sql.ModelBase, sql.ModelDictMixinWithExtras):
    __tablename__ = 'endpoint'
    attributes = [
        'id',
        'interface',
        'region_id',
        'service_id',
        'url',
        'legacy_endpoint_id',
        'enabled',
    ]
    id = sql.Column(sql.String(64), primary_key=True)
    legacy_endpoint_id = sql.Column(sql.String(64))
    interface = sql.Column(sql.String(8), nullable=False)
    region_id = sql.Column(
        sql.String(255),
        sql.ForeignKey('region.id', ondelete='RESTRICT'),
        nullable=True,
        default=None,
    )
    service_id = sql.Column(
        sql.String(64), sql.ForeignKey('service.id'), nullable=False
    )
    url = sql.Column(sql.Text(), nullable=False)
    enabled = sql.Column(
        sql.Boolean,
        nullable=False,
        default=True,
        server_default=sqlalchemy.sql.expression.true(),
    )
    extra = sql.Column(sql.JsonBlob())

    @classmethod
    def from_dict(cls, endpoint_dict):
        """Override from_dict to set enabled if missing."""
        new_dict = endpoint_dict.copy()
        if new_dict.get('enabled') is None:
            new_dict['enabled'] = True
        return super().from_dict(new_dict)


class Catalog(base.CatalogDriverBase):
    # Regions
    def list_regions(self, hints):
        with sql.session_for_read() as session:
            regions = session.query(Region)
            regions = sql.filter_limit_query(Region, regions, hints)
            return [s.to_dict() for s in list(regions)]

    def _get_region(self, session, region_id):
        ref = session.get(Region, region_id)
        if not ref:
            raise exception.RegionNotFound(region_id=region_id)
        return ref

    def _delete_child_regions(self, session, region_id, root_region_id):
        """Delete all child regions.

        Recursively delete any region that has the supplied region
        as its parent.
        """
        children = session.query(Region).filter_by(parent_region_id=region_id)
        for child in children:
            if child.id == root_region_id:
                # Hit a circular region hierarchy
                return
            self._delete_child_regions(session, child.id, root_region_id)
            session.delete(child)

    def _check_parent_region(self, session, region_ref):
        """Raise a NotFound if the parent region does not exist.

        If the region_ref has a specified parent_region_id, check that
        the parent exists, otherwise, raise a NotFound.
        """
        parent_region_id = region_ref.get('parent_region_id')
        if parent_region_id is not None:
            # This will raise NotFound if the parent doesn't exist,
            # which is the behavior we want.
            self._get_region(session, parent_region_id)

    def _has_endpoints(self, session, region, root_region):
        if region.endpoints is not None and len(region.endpoints) > 0:
            return True

        q = session.query(Region)
        q = q.filter_by(parent_region_id=region.id)
        for child in q.all():
            if child.id == root_region.id:
                # Hit a circular region hierarchy
                return False
            if self._has_endpoints(session, child, root_region):
                return True
        return False

    def get_region(self, region_id):
        with sql.session_for_read() as session:
            return self._get_region(session, region_id).to_dict()

    def delete_region(self, region_id):
        with sql.session_for_write() as session:
            ref = self._get_region(session, region_id)
            if self._has_endpoints(session, ref, ref):
                raise exception.RegionDeletionError(region_id=region_id)
            self._delete_child_regions(session, region_id, region_id)
            session.delete(ref)

    @sql.handle_conflicts(conflict_type='region')
    def create_region(self, region_ref):
        with sql.session_for_write() as session:
            self._check_parent_region(session, region_ref)
            region = Region.from_dict(region_ref)
            session.add(region)
            return region.to_dict()

    def update_region(self, region_id, region_ref):
        with sql.session_for_write() as session:
            self._check_parent_region(session, region_ref)
            ref = self._get_region(session, region_id)
            old_dict = ref.to_dict()
            old_dict.update(region_ref)
            self._ensure_no_circle_in_hierarchical_regions(old_dict)
            new_region = Region.from_dict(old_dict)
            for attr in Region.attributes:
                if attr != 'id':
                    setattr(ref, attr, getattr(new_region, attr))
            ref.extra = new_region.extra
            return ref.to_dict()

    # Services
    @driver_hints.truncated
    def list_services(self, hints):
        with sql.session_for_read() as session:
            services = session.query(Service)
            services = sql.filter_limit_query(Service, services, hints)
            return [s.to_dict() for s in list(services)]

    def _get_service(self, session, service_id):
        ref = session.get(Service, service_id)
        if not ref:
            raise exception.ServiceNotFound(service_id=service_id)
        return ref

    def get_service(self, service_id):
        with sql.session_for_read() as session:
            return self._get_service(session, service_id).to_dict()

    def delete_service(self, service_id):
        with sql.session_for_write() as session:
            ref = self._get_service(session, service_id)
            session.query(Endpoint).filter_by(service_id=service_id).delete()
            session.delete(ref)

    def create_service(self, service_id, service_ref):
        with sql.session_for_write() as session:
            service = Service.from_dict(service_ref)
            session.add(service)
            return service.to_dict()

    def update_service(self, service_id, service_ref):
        with sql.session_for_write() as session:
            ref = self._get_service(session, service_id)
            old_dict = ref.to_dict()
            old_dict.update(service_ref)
            new_service = Service.from_dict(old_dict)
            for attr in Service.attributes:
                if attr != 'id':
                    setattr(ref, attr, getattr(new_service, attr))
            ref.extra = new_service.extra
            return ref.to_dict()

    # Endpoints
    def create_endpoint(self, endpoint_id, endpoint):
        with sql.session_for_write() as session:
            endpoint_ref = Endpoint.from_dict(endpoint)
            session.add(endpoint_ref)
            return endpoint_ref.to_dict()

    def delete_endpoint(self, endpoint_id):
        with sql.session_for_write() as session:
            ref = self._get_endpoint(session, endpoint_id)
            session.delete(ref)

    def _get_endpoint(self, session, endpoint_id):
        try:
            return session.query(Endpoint).filter_by(id=endpoint_id).one()
        except sql.NotFound:
            raise exception.EndpointNotFound(endpoint_id=endpoint_id)

    def get_endpoint(self, endpoint_id):
        with sql.session_for_read() as session:
            return self._get_endpoint(session, endpoint_id).to_dict()

    @driver_hints.truncated
    def list_endpoints(self, hints):
        with sql.session_for_read() as session:
            endpoints = session.query(Endpoint)
            endpoints = sql.filter_limit_query(Endpoint, endpoints, hints)
            return [e.to_dict() for e in list(endpoints)]

    def update_endpoint(self, endpoint_id, endpoint_ref):
        with sql.session_for_write() as session:
            ref = self._get_endpoint(session, endpoint_id)
            old_dict = ref.to_dict()
            old_dict.update(endpoint_ref)
            new_endpoint = Endpoint.from_dict(old_dict)
            for attr in Endpoint.attributes:
                if attr != 'id':
                    setattr(ref, attr, getattr(new_endpoint, attr))
            ref.extra = new_endpoint.extra
            return ref.to_dict()

    def get_catalog(self, user_id, project_id):
        """Retrieve and format the V2 service catalog.

        :param user_id: The id of the user who has been authenticated for
            creating service catalog.
        :param project_id: The id of the project. 'project_id' will be None
            in the case this being called to create a catalog to go in a
            domain scoped token. In this case, any endpoint that requires
            a project_id as part of their URL will be skipped (as would a whole
            service if, as a consequence, it has no valid endpoints).

        :returns: A nested dict representing the service catalog or an
                  empty dict.

        """
        substitutions = dict(CONF.items())
        substitutions.update({'user_id': user_id})
        silent_keyerror_failures = []
        if project_id:
            substitutions.update(
                {'tenant_id': project_id, 'project_id': project_id}
            )
        else:
            silent_keyerror_failures = ['tenant_id', 'project_id']

        with sql.session_for_read() as session:
            endpoints = (
                session.query(Endpoint)
                .options(sql.joinedload(Endpoint.service))
                .filter(Endpoint.enabled == true())
                .all()
            )

            catalog = {}

            for endpoint in endpoints:
                if not endpoint.service['enabled']:
                    continue
                try:
                    formatted_url = utils.format_url(
                        endpoint['url'],
                        substitutions,
                        silent_keyerror_failures=silent_keyerror_failures,
                    )
                    if formatted_url is not None:
                        url = formatted_url
                    else:
                        continue
                except exception.MalformedEndpoint:  # nosec(tkelsey)
                    continue  # this failure is already logged in format_url()

                region = endpoint['region_id']
                service_type = endpoint.service['type']
                default_service = {
                    'id': endpoint['id'],
                    'name': endpoint.service.extra.get('name', ''),
                    'publicURL': '',
                }
                catalog.setdefault(region, {})
                catalog[region].setdefault(service_type, default_service)
                interface_url = '{}URL'.format(endpoint['interface'])
                catalog[region][service_type][interface_url] = url

            return catalog

    def get_v3_catalog(self, user_id, project_id):
        """Retrieve and format the current V3 service catalog.

        :param user_id: The id of the user who has been authenticated for
            creating service catalog.
        :param project_id: The id of the project. 'project_id' will be None in
            the case this being called to create a catalog to go in a domain
            scoped token. In this case, any endpoint that requires a
            project_id as part of their URL will be skipped.

        :returns: A list representing the service catalog or an empty list

        """
        d = dict(CONF.items())
        d.update({'user_id': user_id})
        silent_keyerror_failures = []
        if project_id:
            d.update({'tenant_id': project_id, 'project_id': project_id})
        else:
            silent_keyerror_failures = ['tenant_id', 'project_id']

        with sql.session_for_read() as session:
            services = (
                session.query(Service)
                .filter(Service.enabled == true())
                .options(sql.joinedload(Service.endpoints))
                .all()
            )

            def make_v3_endpoints(endpoints):
                for endpoint in (
                    ep.to_dict() for ep in endpoints if ep.enabled
                ):
                    del endpoint['service_id']
                    del endpoint['legacy_endpoint_id']
                    del endpoint['enabled']
                    endpoint['region'] = endpoint['region_id']
                    try:
                        formatted_url = utils.format_url(
                            endpoint['url'],
                            d,
                            silent_keyerror_failures=silent_keyerror_failures,
                        )
                        if formatted_url:
                            endpoint['url'] = formatted_url
                        else:
                            continue
                    except exception.MalformedEndpoint:  # nosec(tkelsey)
                        # this failure is already logged in format_url()
                        continue

                    yield endpoint

            # TODO(davechen): If there is service with no endpoints, we should
            # skip the service instead of keeping it in the catalog,
            # see bug #1436704.
            def make_v3_service(svc):
                eps = list(make_v3_endpoints(svc.endpoints))
                service = {'endpoints': eps, 'id': svc.id, 'type': svc.type}
                service['name'] = svc.extra.get('name', '')
                return service

            # Build the unfiltered catalog, this is the catalog that is
            # returned if endpoint filtering is not performed and the
            # option of `return_all_endpoints_if_no_filter` is set to true.
            catalog_ref = [make_v3_service(svc) for svc in services]

            # Filter the `catalog_ref` above by any project-endpoint
            # association configured by endpoint filter.
            filtered_endpoints = {}
            if project_id:
                filtered_endpoints = (
                    self.catalog_api.list_endpoints_for_project(project_id)
                )
            # endpoint filter is enabled, only return the filtered endpoints.
            if filtered_endpoints:
                filtered_ids = list(filtered_endpoints.keys())
                # This is actually working on the copy of `catalog_ref` since
                # the index will be shifted if remove/add any entry for the
                # original one.
                for service in catalog_ref[:]:
                    endpoints = service['endpoints']
                    for endpoint in endpoints[:]:
                        endpoint_id = endpoint['id']
                        # remove the endpoint that is not associated with
                        # the project.
                        if endpoint_id not in filtered_ids:
                            service['endpoints'].remove(endpoint)
                            continue
                        # remove the disabled endpoint from the list.
                        if not filtered_endpoints[endpoint_id]['enabled']:
                            service['endpoints'].remove(endpoint)
                    # NOTE(davechen): The service will not be included in the
                    # catalog if the service doesn't have any endpoint when
                    # endpoint filter is enabled, this is inconsistent with
                    # full catalog that is returned when endpoint filter is
                    # disabled.
                    if not service.get('endpoints'):
                        catalog_ref.remove(service)
            # When it arrives here it means it's domain scoped token (
            # `project_id` is not set) or it's a project scoped token
            # but the endpoint filtering is not performed.
            # Both of them tell us the endpoint filtering is not enabled, so
            # check the option of `return_all_endpoints_if_no_filter`, it will
            # judge whether a full unfiltered catalog or a empty service
            # catalog will be returned.
            elif not CONF.endpoint_filter.return_all_endpoints_if_no_filter:
                return []
            return catalog_ref

    @sql.handle_conflicts(conflict_type='project_endpoint')
    def add_endpoint_to_project(self, endpoint_id, project_id):
        with sql.session_for_write() as session:
            endpoint_filter_ref = ProjectEndpoint(
                endpoint_id=endpoint_id, project_id=project_id
            )
            session.add(endpoint_filter_ref)

    def _get_project_endpoint_ref(self, session, endpoint_id, project_id):
        endpoint_filter_ref = session.get(
            ProjectEndpoint, (endpoint_id, project_id)
        )
        if endpoint_filter_ref is None:
            msg = _(
                'Endpoint %(endpoint_id)s not found in project '
                '%(project_id)s'
            ) % {'endpoint_id': endpoint_id, 'project_id': project_id}
            raise exception.NotFound(msg)
        return endpoint_filter_ref

    def check_endpoint_in_project(self, endpoint_id, project_id):
        with sql.session_for_read() as session:
            self._get_project_endpoint_ref(session, endpoint_id, project_id)

    def remove_endpoint_from_project(self, endpoint_id, project_id):
        with sql.session_for_write() as session:
            endpoint_filter_ref = self._get_project_endpoint_ref(
                session, endpoint_id, project_id
            )
            session.delete(endpoint_filter_ref)

    def list_endpoints_for_project(self, project_id):
        with sql.session_for_read() as session:
            query = session.query(ProjectEndpoint)
            query = query.filter_by(project_id=project_id)
            endpoint_filter_refs = query.all()
            return [ref.to_dict() for ref in endpoint_filter_refs]

    def list_projects_for_endpoint(self, endpoint_id):
        with sql.session_for_read() as session:
            query = session.query(ProjectEndpoint)
            query = query.filter_by(endpoint_id=endpoint_id)
            endpoint_filter_refs = query.all()
            return [ref.to_dict() for ref in endpoint_filter_refs]

    def delete_association_by_endpoint(self, endpoint_id):
        with sql.session_for_write() as session:
            query = session.query(ProjectEndpoint)
            query = query.filter_by(endpoint_id=endpoint_id)
            query.delete(synchronize_session=False)

    def delete_association_by_project(self, project_id):
        with sql.session_for_write() as session:
            query = session.query(ProjectEndpoint)
            query = query.filter_by(project_id=project_id)
            query.delete(synchronize_session=False)

    def create_endpoint_group(self, endpoint_group_id, endpoint_group):
        with sql.session_for_write() as session:
            endpoint_group_ref = EndpointGroup.from_dict(endpoint_group)
            session.add(endpoint_group_ref)
            return endpoint_group_ref.to_dict()

    def _get_endpoint_group(self, session, endpoint_group_id):
        endpoint_group_ref = session.get(EndpointGroup, endpoint_group_id)
        if endpoint_group_ref is None:
            raise exception.EndpointGroupNotFound(
                endpoint_group_id=endpoint_group_id
            )
        return endpoint_group_ref

    def get_endpoint_group(self, endpoint_group_id):
        with sql.session_for_read() as session:
            endpoint_group_ref = self._get_endpoint_group(
                session, endpoint_group_id
            )
            return endpoint_group_ref.to_dict()

    def update_endpoint_group(self, endpoint_group_id, endpoint_group):
        with sql.session_for_write() as session:
            endpoint_group_ref = self._get_endpoint_group(
                session, endpoint_group_id
            )
            old_endpoint_group = endpoint_group_ref.to_dict()
            old_endpoint_group.update(endpoint_group)
            new_endpoint_group = EndpointGroup.from_dict(old_endpoint_group)
            for attr in EndpointGroup.mutable_attributes:
                setattr(
                    endpoint_group_ref, attr, getattr(new_endpoint_group, attr)
                )
            return endpoint_group_ref.to_dict()

    def delete_endpoint_group(self, endpoint_group_id):
        with sql.session_for_write() as session:
            endpoint_group_ref = self._get_endpoint_group(
                session, endpoint_group_id
            )
            self._delete_endpoint_group_association_by_endpoint_group(
                session, endpoint_group_id
            )
            session.delete(endpoint_group_ref)

    def get_endpoint_group_in_project(self, endpoint_group_id, project_id):
        with sql.session_for_read() as session:
            ref = self._get_endpoint_group_in_project(
                session, endpoint_group_id, project_id
            )
            return ref.to_dict()

    @sql.handle_conflicts(conflict_type='project_endpoint_group')
    def add_endpoint_group_to_project(self, endpoint_group_id, project_id):
        with sql.session_for_write() as session:
            # Create a new Project Endpoint group entity
            endpoint_group_project_ref = ProjectEndpointGroupMembership(
                endpoint_group_id=endpoint_group_id, project_id=project_id
            )
            session.add(endpoint_group_project_ref)

    def _get_endpoint_group_in_project(
        self, session, endpoint_group_id, project_id
    ):
        endpoint_group_project_ref = session.get(
            ProjectEndpointGroupMembership, (endpoint_group_id, project_id)
        )
        if endpoint_group_project_ref is None:
            msg = _('Endpoint Group Project Association not found')
            raise exception.NotFound(msg)
        else:
            return endpoint_group_project_ref

    def list_endpoint_groups(self, hints):
        with sql.session_for_read() as session:
            query = session.query(EndpointGroup)
            endpoint_group_refs = sql.filter_limit_query(
                EndpointGroup, query, hints
            )
            return [e.to_dict() for e in endpoint_group_refs]

    def list_endpoint_groups_for_project(self, project_id):
        with sql.session_for_read() as session:
            query = session.query(ProjectEndpointGroupMembership)
            query = query.filter_by(project_id=project_id)
            endpoint_group_refs = query.all()
            return [ref.to_dict() for ref in endpoint_group_refs]

    def remove_endpoint_group_from_project(
        self, endpoint_group_id, project_id
    ):
        with sql.session_for_write() as session:
            endpoint_group_project_ref = self._get_endpoint_group_in_project(
                session, endpoint_group_id, project_id
            )
            session.delete(endpoint_group_project_ref)

    def list_projects_associated_with_endpoint_group(self, endpoint_group_id):
        with sql.session_for_read() as session:
            query = session.query(ProjectEndpointGroupMembership)
            query = query.filter_by(endpoint_group_id=endpoint_group_id)
            endpoint_group_refs = query.all()
            return [ref.to_dict() for ref in endpoint_group_refs]

    def _delete_endpoint_group_association_by_endpoint_group(
        self, session, endpoint_group_id
    ):
        query = session.query(ProjectEndpointGroupMembership)
        query = query.filter_by(endpoint_group_id=endpoint_group_id)
        query.delete()

    def delete_endpoint_group_association_by_project(self, project_id):
        with sql.session_for_write() as session:
            query = session.query(ProjectEndpointGroupMembership)
            query = query.filter_by(project_id=project_id)
            query.delete()


class ProjectEndpoint(sql.ModelBase, sql.ModelDictMixin):
    """project-endpoint relationship table."""

    __tablename__ = 'project_endpoint'
    attributes = ['endpoint_id', 'project_id']
    endpoint_id = sql.Column(sql.String(64), primary_key=True, nullable=False)
    project_id = sql.Column(sql.String(64), primary_key=True, nullable=False)


class EndpointGroup(sql.ModelBase, sql.ModelDictMixin):
    """Endpoint Groups table."""

    __tablename__ = 'endpoint_group'
    attributes = ['id', 'name', 'description', 'filters']
    mutable_attributes = frozenset(['name', 'description', 'filters'])
    id = sql.Column(sql.String(64), primary_key=True)
    name = sql.Column(sql.String(255), nullable=False)
    description = sql.Column(sql.Text, nullable=True)
    filters = sql.Column(sql.JsonBlob(), nullable=False)


class ProjectEndpointGroupMembership(sql.ModelBase, sql.ModelDictMixin):
    """Project to Endpoint group relationship table."""

    __tablename__ = 'project_endpoint_group'
    attributes = ['endpoint_group_id', 'project_id']
    endpoint_group_id = sql.Column(
        sql.String(64), sql.ForeignKey('endpoint_group.id'), nullable=False
    )
    project_id = sql.Column(sql.String(64), nullable=False)
    __table_args__ = (
        sql.PrimaryKeyConstraint('endpoint_group_id', 'project_id'),
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystone.limit.models import base


class FlatModel(base.ModelBase):
    NAME = 'flat'
    DESCRIPTION = (
        'Limit enforcement and validation does not take project hierarchy '
        'into consideration.'
    )
    MAX_PROJECT_TREE_DEPTH = None

    def check_limit(self, limits):
        # Flat limit model is not hierarchical, so don't need to check the
        # value.
        return



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
import abc
import binascii
import typing as ty

# Methods as implemented in passlib to guarantee backwards compatibility.

_BASE64_STRIP = b"=\n"
_BASE64_PAD1 = b"="
_BASE64_PAD2 = b"=="


def b64s_decode(data: bytes):
    """decode from shortened base64 format which omits padding & whitespace

    uses default ``+/`` altchars.
    """
    off = len(data) & 3
    if off == 0:
        pass
    elif off == 2:
        data += _BASE64_PAD2
    elif off == 3:
        data += _BASE64_PAD1
    else:  # off == 1
        raise ValueError("invalid base64 input")
    return binascii.a2b_base64(data)


class PasswordHasher(abc.ABC):
    """Abstract password hasher class"""

    name: str
    ident: ty.Optional[str]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re

import keystone.conf

CONF = keystone.conf.CONF


def symptom_minimum_password_age_greater_than_expires_days():
    """Minimum password age should be less than the password expires days.

    If the minimum password age is greater than or equal to the password
    expires days, then users would not be able to change their passwords before
    they expire.

    Ensure `[security_compliance] minimum_password_age` is less than the
    `[security_compliance] password_expires_days`.
    """
    min_age = CONF.security_compliance.minimum_password_age
    expires = CONF.security_compliance.password_expires_days
    return (min_age >= expires) if (min_age > 0 and expires > 0) else False


def symptom_invalid_password_regular_expression():
    """Invalid password regular expression.

    The password regular expression is invalid and users will not be able to
    make password changes until this has been corrected.

    Ensure `[security_compliance] password_regex` is a valid regular
    expression.
    """
    try:
        if CONF.security_compliance.password_regex:
            re.match(CONF.security_compliance.password_regex, 'password')
        return False
    except re.error:
        return True


def symptom_password_regular_expression_description_not_set():
    """Password regular expression description is not set.

    The password regular expression is set, but the description is not. Thus,
    if a user fails the password regular expression, they will not receive a
    message to explain why their requested password was insufficient.

    Ensure `[security_compliance] password_regex_description` is set with a
    description of your password regular expression in a language for humans.
    """
    return (
        CONF.security_compliance.password_regex
        and not CONF.security_compliance.password_regex_description
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add Identity Federation attribute mapping schema version.

Revision ID: 47147121
Revises: 11c3b243b4cb
Create Date: 2023-12-12 09:00:00
"""

from alembic import op
from sqlalchemy import Column
from sqlalchemy import String

# revision identifiers, used by Alembic.
revision = '47147121'
down_revision = '11c3b243b4cb'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "mapping",
        Column(
            'schema_version', String(5), nullable=False, server_default="1.0"
        ),
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import keystone.conf

CONF = keystone.conf.CONF


def symptom_debug_mode_is_enabled():
    """Debug mode should be set to False.

    Debug mode can be used to get more information back when trying to isolate
    a problem, but it is not recommended to be enabled when running a
    production environment.

    Ensure `keystone.conf debug` is set to False
    """
    return CONF.debug



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Common field types for validating API responses."""

from typing import Any

# Common schema for resource `link` attribute
links: dict[str, Any] = {
    "type": "object",
    "description": "Links for the collection of resources.",
    "properties": {
        "next": {"type": ["string", "null"], "format": "uri"},
        "previous": {"type": ["string", "null"], "format": "uri"},
        "self": {"type": "string", "format": "uri"},
    },
    "required": ["self"],
    "additionalProperties": False,
    "readOnly": True,
}

# Resource `links` attribute schema
resource_links: dict[str, Any] = {
    "type": "object",
    "description": "The link to the resource in question.",
    "properties": {"self": {"type": "string", "format": "uri"}},
    "additionalProperties": False,
    "readOnly": True,
}

truncated: dict[str, Any] = {
    "type": "boolean",
    "description": (
        "Flag indicating that the amount of entities exceeds global "
        "response limit"
    ),
}

tags: dict[str, Any] = {
    "type": "object",
    "properties": {
        "tags": {"type": "array", "items": {"type": "string"}},
        "links": links,
    },
    "additionalProperties": False,
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from keystone.server.flask import core as flask_core


# NOTE(morgan): While "_get_config_files" is present in the keystone_flask
# module, since it is considered "private", we are going to directly
# import core and call it directly, eventually keystone_flask will not
# export all the symbols from keystone.flask.core only specific ones that
# are meant for public consumption
def initialize_public_application():
    return flask_core.initialize_application(
        name='public', config_files=flask_core._get_config_files()
    )


# Keystone does not differentiate between "admin" and public with the removal
# of V2.0
initialize_admin_application = initialize_public_application



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import constants
from keystone.conf import utils

assertion_expiration_time = cfg.IntOpt(
    'assertion_expiration_time',
    default=3600,
    help=utils.fmt(
        """
Determines the lifetime for any SAML assertions generated by keystone, using
`NotOnOrAfter` attributes.
"""
    ),
)

xmlsec1_binary = cfg.StrOpt(
    'xmlsec1_binary',
    default='xmlsec1',
    help=utils.fmt(
        """
Name of, or absolute path to, the binary to be used for XML signing. Although
only the XML Security Library (`xmlsec1`) is supported, it may have a
non-standard name or path on your system. If keystone cannot find the binary
itself, you may need to install the appropriate package, use this option to
specify an absolute path, or adjust keystone's PATH environment variable.
"""
    ),
)

certfile = cfg.StrOpt(
    'certfile',
    default=constants._CERTFILE,
    help=utils.fmt(
        """
Absolute path to the public certificate file to use for SAML signing. The value
cannot contain a comma (`,`).
"""
    ),
)

keyfile = cfg.StrOpt(
    'keyfile',
    default=constants._KEYFILE,
    help=utils.fmt(
        """
Absolute path to the private key file to use for SAML signing. The value cannot
contain a comma (`,`).
"""
    ),
)

idp_entity_id = cfg.URIOpt(
    'idp_entity_id',
    max_length=1024,
    help=utils.fmt(
        """
This is the unique entity identifier of the identity provider (keystone) to use
when generating SAML assertions. This value is required to generate identity
provider metadata and must be a URI (a URL is recommended). For example:
`https://keystone.example.com/v3/OS-FEDERATION/saml2/idp`.
"""
    ),
)

idp_sso_endpoint = cfg.URIOpt(
    'idp_sso_endpoint',
    help=utils.fmt(
        """
This is the single sign-on (SSO) service location of the identity provider
which accepts HTTP POST requests. A value is required to generate identity
provider metadata. For example:
`https://keystone.example.com/v3/OS-FEDERATION/saml2/sso`.
"""
    ),
)

idp_lang = cfg.StrOpt(
    'idp_lang',
    default='en',
    help=utils.fmt(
        """
This is the language used by the identity provider's organization.
"""
    ),
)

idp_organization_name = cfg.StrOpt(
    'idp_organization_name',
    default='SAML Identity Provider',
    help=utils.fmt(
        """
This is the name of the identity provider's organization.
"""
    ),
)

idp_organization_display_name = cfg.StrOpt(
    'idp_organization_display_name',
    default='OpenStack SAML Identity Provider',
    help=utils.fmt(
        """
This is the name of the identity provider's organization to be displayed.
"""
    ),
)

idp_organization_url = cfg.URIOpt(
    'idp_organization_url',
    default='https://example.com/',
    help=utils.fmt(
        """
This is the URL of the identity provider's organization. The URL referenced
here should be useful to humans.
"""
    ),
)

idp_contact_company = cfg.StrOpt(
    'idp_contact_company',
    default='Example, Inc.',
    help=utils.fmt(
        """
This is the company name of the identity provider's contact person.
"""
    ),
)

idp_contact_name = cfg.StrOpt(
    'idp_contact_name',
    default='SAML Identity Provider Support',
    help=utils.fmt(
        """
This is the given name of the identity provider's contact person.
"""
    ),
)

idp_contact_surname = cfg.StrOpt(
    'idp_contact_surname',
    default='Support',
    help=utils.fmt(
        """
This is the surname of the identity provider's contact person.
"""
    ),
)

idp_contact_email = cfg.StrOpt(
    'idp_contact_email',
    default='support@example.com',
    help=utils.fmt(
        """
This is the email address of the identity provider's contact person.
"""
    ),
)

idp_contact_telephone = cfg.StrOpt(
    'idp_contact_telephone',
    default='+1 800 555 0100',
    help=utils.fmt(
        """
This is the telephone number of the identity provider's contact person.
"""
    ),
)

idp_contact_type = cfg.StrOpt(
    'idp_contact_type',
    default='other',
    choices=['technical', 'support', 'administrative', 'billing', 'other'],
    help=utils.fmt(
        """
This is the type of contact that best describes the identity provider's contact
person.
"""
    ),
)

idp_metadata_path = cfg.StrOpt(
    'idp_metadata_path',
    default='/etc/keystone/saml2_idp_metadata.xml',
    help=utils.fmt(
        """
Absolute path to the identity provider metadata file. This file should be
generated with the `keystone-manage saml_idp_metadata` command. There is
typically no reason to change this value.
"""
    ),
)

relay_state_prefix = cfg.StrOpt(
    'relay_state_prefix',
    default='ss:mem:',
    help=utils.fmt(
        """
The prefix of the RelayState SAML attribute to use when generating enhanced
client and proxy (ECP) assertions. In a typical deployment, there is no reason
to change this value.
"""
    ),
)


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [
    assertion_expiration_time,
    xmlsec1_binary,
    certfile,
    keyfile,
    idp_entity_id,
    idp_sso_endpoint,
    idp_lang,
    idp_organization_name,
    idp_organization_display_name,
    idp_organization_url,
    idp_contact_company,
    idp_contact_name,
    idp_contact_surname,
    idp_contact_email,
    idp_contact_telephone,
    idp_contact_type,
    idp_metadata_path,
    relay_state_prefix,
]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""API request/response validating middleware."""

import functools
import typing as ty

import flask
from oslo_serialization import jsonutils

from keystone.api.validation import validators


def validated(cls):
    cls._validated = True
    return cls


def _schema_validator(
    schema: ty.Dict[str, ty.Any],
    target: ty.Dict[str, ty.Any],
    args: ty.Any,
    kwargs: ty.Any,
    is_body: bool = True,
):
    """A helper method to execute JSON Schema Validation.

    This method checks the request version whether matches the specified
    ``max_version`` and ``min_version``. If the version range matches the
    request, we validate ``schema`` against ``target``. A failure will result
    in ``ValidationError`` being raised.

    :param schema: The JSON Schema schema used to validate the target.
    :param target: The target to be validated by the schema.
    :param args: Positional arguments which passed into original method.
    :param kwargs: Keyword arguments which passed into original method.
    :param is_body: Whether ``target`` is a HTTP request body or not.
    :returns: None.
    :raises: ``ValidationError`` if validation fails.
    """
    schema_validator = validators._SchemaValidator(schema, is_body=is_body)
    schema_validator.validate(target)


def request_body_schema(schema: ty.Optional[ty.Dict[str, ty.Any]] = None):
    """Register a schema to validate request body.

    ``schema`` will be used for validating the request body just before the API
    method is executed.

    :param schema: The JSON Schema schema used to validate the target. If
        empty value is passed no validation will be performed.
    :param min_version: A string indicating the minimum API version ``schema``
        applies against.
    :param max_version: A string indicating the maximum API version ``schema``
        applies against.
    """

    def add_validator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if schema is not None:
                _schema_validator(
                    schema,
                    flask.request.get_json(silent=True, force=True) or {},
                    args,
                    kwargs,
                    is_body=True,
                )
            return func(*args, **kwargs)

        wrapper._request_body_schema = schema

        return wrapper

    return add_validator


def request_query_schema(schema: ty.Optional[ty.Dict[str, ty.Any]] = None):
    """Register a schema to validate request query string parameters.

    ``schema`` will be used for validating request query strings just before
    the API method is executed.

    :param schema: The JSON Schema schema used to validate the target. If
        empty value is passed no validation will be performed.
    """

    def add_validator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if schema is not None:
                # NOTE: The request object is always the second argument.
                # However, numerous unittests pass in the request object
                # via kwargs instead so we handle that as well.
                # TODO(stephenfin): Fix unit tests so we don't have to do this
                if "req" in kwargs:
                    req = kwargs["req"]
                else:
                    req = flask.request.args

                _schema_validator(schema, req, args, kwargs, is_body=True)
            return func(*args, **kwargs)

        wrapper._request_query_schema = schema

        return wrapper

    return add_validator


def response_body_schema(schema: ty.Optional[ty.Dict[str, ty.Any]] = None):
    """Register a schema to validate response body.

    ``schema`` will be used for validating the response body just after the API
    method is executed.

    :param schema: The JSON Schema schema used to validate the target. If
        empty value is passed no validation will be performed.
    :param min_version: A string indicating the minimum API version ``schema``
        applies against.
    :param max_version: A string indicating the maximum API version ``schema``
        applies against.
    """

    def add_validator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            response = func(*args, **kwargs)

            if schema is not None:
                # In Flask it is not uncommon that the method return a tuple of
                # body and the status code. In the runtime Keystone only return
                # a body, but some of the used testtools do return a tuple.
                if isinstance(response, tuple):
                    _body = response[0]
                else:
                    _body = response

                # NOTE(stephenfin): If our response is an object, we need to
                # serializer and deserialize to convert e.g. date-time
                # to strings
                _body = jsonutils.dump_as_bytes(_body)

                if _body == b"":
                    body = None
                else:
                    body = jsonutils.loads(_body)

                _schema_validator(schema, body, args, kwargs, is_body=True)
            return response

        wrapper._response_body_schema = schema

        return wrapper

    return add_validator



# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import hashlib
import http.client

from keystoneclient.contrib.ec2 import utils as ec2_utils
from oslo_utils import timeutils

from keystone.common import provider_api
from keystone.common import utils
from keystone.tests import unit
from keystone.tests.unit import test_v3

PROVIDERS = provider_api.ProviderAPIs


class EC2ContribCoreV3(test_v3.RestfulTestCase):
    def setUp(self):
        super().setUp()

        self.cred_blob, self.credential = unit.new_ec2_credential(
            self.user['id'], self.project_id
        )
        PROVIDERS.credential_api.create_credential(
            self.credential['id'], self.credential
        )

    def test_http_get_method_not_allowed(self):
        resp = self.get(
            '/ec2tokens',
            expected_status=http.client.METHOD_NOT_ALLOWED,
            convert=False,
        )
        self.assertEqual(http.client.METHOD_NOT_ALLOWED, resp.status_code)

    def _test_valid_authentication_response_with_proper_secret(self, **kwargs):
        signer = ec2_utils.Ec2Signer(self.cred_blob['secret'])
        timestamp = utils.isotime(timeutils.utcnow())
        credentials = {
            'access': self.cred_blob['access'],
            'secret': self.cred_blob['secret'],
            'host': 'localhost',
            'verb': 'GET',
            'path': '/',
            'params': {
                'SignatureVersion': '2',
                'Action': 'Test',
                'Timestamp': timestamp,
            },
        }
        credentials['signature'] = signer.generate(credentials)
        resp = self.post(
            '/ec2tokens',
            body={'credentials': credentials},
            expected_status=http.client.OK,
            **kwargs,
        )
        self.assertValidProjectScopedTokenResponse(resp, self.user)

    def test_valid_authentication_response_with_proper_secret(self):
        self._test_valid_authentication_response_with_proper_secret()

    def test_valid_authentication_response_with_proper_secret_noauth(self):
        self._test_valid_authentication_response_with_proper_secret(
            noauth=True
        )

    def test_valid_authentication_response_with_signature_v4(self):
        signer = ec2_utils.Ec2Signer(self.cred_blob['secret'])
        timestamp = utils.isotime(timeutils.utcnow())
        hashed_payload = (
            'GET\n'
            '/\n'
            'Action=Test\n'
            'host:localhost\n'
            'x-amz-date:' + timestamp + '\n'
            '\n'
            'host;x-amz-date\n'
            'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
        )
        body_hash = hashlib.sha256(hashed_payload.encode()).hexdigest()
        amz_credential = (
            f'AKIAIOSFODNN7EXAMPLE/{timestamp[:8]}/us-east-1/iam/aws4_request,'
        )

        credentials = {
            'access': self.cred_blob['access'],
            'secret': self.cred_blob['secret'],
            'host': 'localhost',
            'verb': 'GET',
            'path': '/',
            'params': {
                'Action': 'Test',
                'X-Amz-Algorithm': 'AWS4-HMAC-SHA256',
                'X-Amz-SignedHeaders': 'host,x-amz-date,',
                'X-Amz-Credential': amz_credential,
            },
            'headers': {'X-Amz-Date': timestamp},
            'body_hash': body_hash,
        }
        credentials['signature'] = signer.generate(credentials)
        resp = self.post(
            '/ec2tokens',
            body={'credentials': credentials},
            expected_status=http.client.OK,
        )
        self.assertValidProjectScopedTokenResponse(resp, self.user)

    def test_authenticate_with_empty_body_returns_bad_request(self):
        self.post(
            '/ec2tokens', body={}, expected_status=http.client.BAD_REQUEST
        )

    def test_authenticate_without_json_request_returns_bad_request(self):
        self.post(
            '/ec2tokens',
            body='not json',
            expected_status=http.client.BAD_REQUEST,
        )

    def test_authenticate_without_request_body_returns_bad_request(self):
        self.post('/ec2tokens', expected_status=http.client.BAD_REQUEST)

    def test_authenticate_without_proper_secret_returns_unauthorized(self):
        signer = ec2_utils.Ec2Signer('totally not the secret')
        timestamp = utils.isotime(timeutils.utcnow())
        credentials = {
            'access': self.cred_blob['access'],
            'secret': 'totally not the secret',
            'host': 'localhost',
            'verb': 'GET',
            'path': '/',
            'params': {
                'SignatureVersion': '2',
                'Action': 'Test',
                'Timestamp': timestamp,
            },
        }
        credentials['signature'] = signer.generate(credentials)
        self.post(
            '/ec2tokens',
            body={'credentials': credentials},
            expected_status=http.client.UNAUTHORIZED,
        )

    def test_authenticate_expired_request(self):
        self.config_fixture.config(group='credential', auth_ttl=5)
        signer = ec2_utils.Ec2Signer(self.cred_blob['secret'])
        past = timeutils.utcnow() - datetime.timedelta(minutes=10)
        timestamp = utils.isotime(past)
        credentials = {
            'access': self.cred_blob['access'],
            'secret': self.cred_blob['secret'],
            'host': 'localhost',
            'verb': 'GET',
            'path': '/',
            'params': {
                'SignatureVersion': '2',
                'Action': 'Test',
                'Timestamp': timestamp,
            },
        }
        credentials['signature'] = signer.generate(credentials)
        self.post(
            '/ec2tokens',
            body={'credentials': credentials},
            expected_status=http.client.UNAUTHORIZED,
        )

    def test_authenticate_expired_request_v4(self):
        self.config_fixture.config(group='credential', auth_ttl=5)
        signer = ec2_utils.Ec2Signer(self.cred_blob['secret'])
        past = timeutils.utcnow() - datetime.timedelta(minutes=10)
        timestamp = utils.isotime(past)
        hashed_payload = (
            'GET\n'
            '/\n'
            'Action=Test\n'
            'host:localhost\n'
            'x-amz-date:' + timestamp + '\n'
            '\n'
            'host;x-amz-date\n'
            'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
        )
        body_hash = hashlib.sha256(hashed_payload.encode()).hexdigest()
        amz_credential = (
            f'AKIAIOSFODNN7EXAMPLE/{timestamp[:8]}/us-east-1/iam/aws4_request,'
        )

        credentials = {
            'access': self.cred_blob['access'],
            'secret': self.cred_blob['secret'],
            'host': 'localhost',
            'verb': 'GET',
            'path': '/',
            'params': {
                'Action': 'Test',
                'X-Amz-Algorithm': 'AWS4-HMAC-SHA256',
                'X-Amz-SignedHeaders': 'host,x-amz-date,',
                'X-Amz-Credential': amz_credential,
            },
            'headers': {'X-Amz-Date': timestamp},
            'body_hash': body_hash,
        }
        credentials['signature'] = signer.generate(credentials)
        self.post(
            '/ec2tokens',
            body={'credentials': credentials},
            expected_status=http.client.UNAUTHORIZED,
        )
